/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.editor;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eclipse.core.runtime.Assert;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.forms.IManagedForm;

import eu.etaxonomy.cdm.model.name.HomotypicalGroup;
import eu.etaxonomy.cdm.model.taxon.Synonym;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationship;
import eu.etaxonomy.taxeditor.controller.EditorController;
import eu.etaxonomy.taxeditor.editor.name.AcceptedNameComposite;
import eu.etaxonomy.taxeditor.editor.name.ConceptComposite;
import eu.etaxonomy.taxeditor.editor.name.ConceptGroupComposite;
import eu.etaxonomy.taxeditor.editor.name.HomotypicalGroupComposite;
import eu.etaxonomy.taxeditor.editor.name.MisappliedGroupComposite;
import eu.etaxonomy.taxeditor.editor.name.MisappliedNameComposite;
import eu.etaxonomy.taxeditor.editor.name.SynonymComposite;

/**
 * @author p.ciardelli
 * @created 08.01.2009
 * @version 1.0
 */
public class FreeTextElementFactory implements IFreeTextElementFactory {
	private static final Logger logger = Logger
			.getLogger(FreeTextElementFactory.class);
	
	private static IFreeTextElementFactory factory;
	
	private Map<Taxon, IManagedForm> nameForms;
	

	/**
	 * Returns a taxon's composite which holds all misapplied names.
	 */
	private Map<Taxon, Composite> misappliedComposites;
	
	public static IFreeTextElementFactory getDefault() {
		if (factory == null) {
			factory = new FreeTextElementFactory();
		}
		return factory;
	}
	
	public Composite createAcceptedTaxon(IManagedForm form, Taxon taxon) {
		
		// Put the taxon's form in the form map
		Assert.isNotNull(form, "A non-null IManagedForm must be provided.");
		getTaxonForms().put(taxon, form);
		
		// Create a homotypic group composite for the accepted taxon
		HomotypicalGroup group = taxon.getHomotypicGroup();
		Assert.isNotNull(group, "Taxon does not have a homotypic group");
		Composite groupComposite = createHomotypicalGroup(taxon, group);
				
		// Create a name composite for the accepted taxon
		Composite acceptedNameComposite =
				new AcceptedNameComposite(groupComposite, form, taxon);
		
	
		// Return the name composite
		return acceptedNameComposite;
	}
	
	public Composite createSynonym(Taxon taxon, Synonym synonym) {
		
		// Get the taxon's form
		IManagedForm form = getForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the synonym's accepted taxon.");
		
		// Get the synonym's homotypic group
		HomotypicalGroup group = synonym.getHomotypicGroup();
		Assert.isNotNull(group, "Synonym does not have a homotypic group");
		
		// If the group doesn't yet have a composite, create one and add it to the repository
		Composite groupComposite = 
				EditorController.getEditor(taxon).getHomotypicGroup(group);
		if (groupComposite == null) {
			groupComposite = createHomotypicalGroup(taxon, group);
		}
				
		// Create a synonym composite in the homotypic group
		Composite synonymComposite = 
				new SynonymComposite(groupComposite, form, taxon, synonym);
				
		return synonymComposite;
	}
	
	@Override
	public Composite createMisappliedName(Taxon taxon, Taxon misappliedName) {

		// Get the taxon's form
		IManagedForm form = getForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the synonym's accepted taxon.");
				
		// If there is no composite for misapplied names, 
		//	create one and add it to the repository
		Composite groupComposite = 
				EditorController.getEditor(taxon).getMisappliedGroup();
		if (groupComposite == null) {
			groupComposite = createMisappliedGroup(taxon);
		}
		
		// Create the name's composite
		Composite composite = new MisappliedNameComposite
				(groupComposite, form, taxon, misappliedName);
		
		return composite;
	}
	
	
	public Composite createMisappliedGroup(Taxon taxon) {
		
		// Get the taxon's form
		IManagedForm form = getForm(taxon);

		// Create the group composite
		Composite groupComposite = new MisappliedGroupComposite(getTopLevelComposite(form), taxon);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));
		
		// Put the group composite before concept group composite, if any
		Composite conceptGroupComposite = 
			EditorController.getEditor(taxon).getConceptGroup();
		if (conceptGroupComposite != null) {
			groupComposite.moveAbove(conceptGroupComposite);
		}
		
		return groupComposite;
	}
	
	public Composite createHomotypicalGroup(Taxon taxon, HomotypicalGroup group) {

		// Get the taxon's form
		IManagedForm form = getForm(taxon);

		// Create the group composite
		Composite groupComposite = new HomotypicalGroupComposite(getTopLevelComposite(form), taxon, group);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));		
				
		return groupComposite;
	}

	private Map<Taxon, IManagedForm> getTaxonForms() {
		if (nameForms == null) {
			nameForms = new HashMap<Taxon, IManagedForm>();
		}
		return nameForms;
	}
	
	private IManagedForm getForm(Taxon taxon) {
		return getTaxonForms().get(taxon);
	}
		
	private Composite getTopLevelComposite(IManagedForm form) {
		return form.getForm().getBody();
	}

	@Override
	public Composite createConcept(Taxon taxon, TaxonRelationship relationship) {
		// Get the taxon's form
		IManagedForm form = getForm(taxon);
		Assert.isNotNull(form, "No IManagedForm has been provided for the synonym's accepted taxon.");
				
		// If there is no composite for misapplied names, 
		//	create one and add it to the repository
		Composite groupComposite = 
				EditorController.getEditor(taxon).getConceptGroup();
		if (groupComposite == null) {
			groupComposite = createConceptGroup(taxon);
		}
		
		// Create the name's composite
		Composite composite = ConceptComposite.getNewInstance(groupComposite, form, 
									taxon, relationship);
				
		return composite;
	}

	@Override
	public Composite createConceptGroup(Taxon taxon) {
		
		// Get the taxon's form
		IManagedForm form = getForm(taxon);

		// Create the group composite
		Composite groupComposite = new ConceptGroupComposite(getTopLevelComposite(form), taxon);
		groupComposite.addFocusListener(new CompositeBorderDecorator(groupComposite, form));
		
		// Put the group composite after misapplied group composite, if any
		Composite misappliedGroupComposite = 
			EditorController.getEditor(taxon).getMisappliedGroup();
		if (misappliedGroupComposite != null) {
			groupComposite.moveBelow(misappliedGroupComposite);
		}
		
		return groupComposite;

	}
}
