/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.model;

import java.util.Collection;
import java.util.Set;
import java.util.SortedSet;

import org.eclipse.core.databinding.observable.set.IObservableSet;

import eu.etaxonomy.cdm.api.application.CdmApplicationController;
import eu.etaxonomy.cdm.model.common.OrderedTermVocabulary;
import eu.etaxonomy.cdm.model.common.TermVocabulary;
import eu.etaxonomy.cdm.model.description.Feature;
import eu.etaxonomy.cdm.model.name.NameRelationshipType;
import eu.etaxonomy.cdm.model.name.NomenclaturalStatusType;
import eu.etaxonomy.cdm.model.name.Rank;
import eu.etaxonomy.cdm.model.reference.ReferenceBase;
import eu.etaxonomy.cdm.model.taxon.Taxon;
import eu.etaxonomy.cdm.model.taxon.TaxonRelationshipType;

/**
 * All session data - taxa, term vocabularies, etc. - are stored in a
 * session data repository. Any CDM data that is stored locally during execution
 * should be kept here while the Editor is running. This makes it convenient
 * to clear all data in case everything needs to be reloaded, for example, when
 * the data source is changed. 
 * <p>
 * The repository uses a CDM application controller to retrieve data.
 * <p>
 * Any method starting with "add" or "remove" deals with data solely in the session.
 * For instance, <code>addTaxon(taxon)</code> will cause a new <code>taxon</code>
 * to appear in the taxonomic tree, but will not save it to the CDM data source.
 * <p>
 * Any method starting with "save" or "delete" will call its corresponding "add"/"remove"
 * method, and commit the change to the CDM data source.  
 * 
 * @author p.ciardelli
 * @created 16.12.2008
 * @version 1.0
 */
public interface ICdmSessionDataRepository {

	/**
	 * Returns the root taxa associated with the current data source, and
	 * adds them to session data if not already there.
	 *  
	 * @return
	 */
	public Set<Taxon> getRootTaxa();
	
	/**
	 * Returns all of a taxon's children, and adds them to session data
	 * if not already there.
	 *  
	 * @param parentTaxon
	 * @return
	 */
	public Set<Taxon> getTaxonomicChildren(Taxon parentTaxon);
	
	/**
	 * Get the default sec. reference used by the current dataset
	 * @return
	 */
	public ReferenceBase getDefaultSec();
	
	/**
	 * Adds a taxon to the repository.
	 * 
	 * @param taxon
	 */
	public void addTaxon(Taxon taxon);

	/**
	 * Removes a taxon from the repository.
	 * 
	 * @param taxon
	 */
	public void removeTaxon(Taxon taxon);

	/**
	 * Removes all taxa from the repository.
	 * replaces clearSessionTaxa
	 * 
	 * @param taxon
	 */
	public void removeAllTaxa();
	
	/**
	 * This is a set that can be used to be notified when there are changes
	 * to the list of all taxa, or to their names, for instance, when the name
	 * of a taxon displayed in a tree changes. 
	 * 
	 * @return
	 */
	public IObservableSet getObservableTaxa();
	
	public void setApplicationController(CdmApplicationController applicationController);
	
	/**
	 * Sets all term vocabularies to NULL to be regenerated the next time
	 * they are needed, i.e. when a new transaction is opened.
	 */
	public void clearNonTaxonData();

	/**
	 * Sets all taxon collections to NULL to be regenerated the next time
	 * they are needed.
	 */
	public void clearTaxonData();
	
	public void clearAllData();
	
	public void addTaxonSetListener(ICdmTaxonSetListener listener);

	public Collection<Taxon> getAllTaxa();	
	
	public boolean saveTaxon(Taxon taxon);
	
	public boolean deleteTaxon(Taxon taxon);

	public void setTaxonomicParent(Taxon taxon, Taxon newParentTaxon);
	
	public TermVocabulary<Feature> getFeatures();
	
	public SortedSet<Rank> getRanks();
	
	public TermVocabulary<NomenclaturalStatusType> getNomStatus();
	
	public SortedSet<NameRelationshipType> getNameRelationshipTypes();

	public OrderedTermVocabulary<TaxonRelationshipType> getTaxonRelationshipTypes();
	
	public Set<TaxonRelationshipType> getConceptRelationshipTypes();
}
